# TODO:
# - genericCopy: content must be copied, too

import xml.dom.minidom
import copy

True = 1 # LEGACY
False = 0 # LEGACY

class TubeEvalException(Exception):
    def __init__(self, *args):
        self.value = args
        
    def __str__(self):
        return repr(self.value)

class TubeFormatException(Exception):
    def __init__(self, *args):
        self.value = args
        
    def __str__(self):
        return repr(self.value)


class TubeEdge:
    
    def initFromData(self, type, shared, target, name=None):
        """
        - type: on of either ["parent","child","delegate"]
        """
        if type not in ["parent","child","delegate"]:
            raise ValueError("unkown type: " + type)
        self.name = name
        self.target = target
        self.type = type
        self.shared = shared
        return self
    
    def initFromXML(self, xmlNode, idDict):
        name     = xmlNode.getAttribute("name")
        targetID = xmlNode.getAttribute("target")

        typeNode = xmlNode.getElementsByTagName("type")[0]
        type = typeNode.getAttribute("kind")

        shared = "shared" in typeNode.attributes.keys()
        target = idDict[targetID]

        self.initFromData(type= type, shared= shared, name=name, target=target)
        return self

    def genericCopy(self, followEdge, copyDeeply):
        """
        - Constructing a node this way does not call the constructor
        """
        myCopy = copy.copy(self) # shallow copy
        if copyDeeply(self):
            myCopy.target = self.target.genericCopy(followEdge=followEdge,
                                                    copyDeeply=copyDeeply)
        return myCopy

    #------------

    def pointsUp(self):
        return self.type in ["parent","delegate"]

    def pointsDown(self):
        return self.type == "child"

    def isShared(self):
        return self.shared

    def isParent(self):
        return self.type == "parent"

    def isDelegate(self):
        return self.type == "delegate"

    def toXML(self):
        if self.name == None:
            name = ''
        else:
            name = ' name="%s"' % self.name
        type     = self.type
        targetID = self.target._id
        
        if self.shared:
            shared = 'shared="true"'
        else:
            shared = ''
            
        return """
            <edge%(name)s target="%(targetID)s">
                <type kind="%(type)s" %(shared)s />
            </edge>
        """ % vars() # a dictionary that contains all local variables

class TubeContent:
    def __init__(self):
        "Create empty content"
        self.calledEdge = None
        self.code = None
        self.codeStr = None

    def initFromData(self, code, kind):
        self.code = code
        self.kind = kind
        return self
    
    def initFromXML(self, xmlNode):
        # For better output: strip trailing whitespace
        codeStr = xmlNode.firstChild.data.rstrip() # text inside the element

        self.kind = xmlNode.getAttribute("kind")
        if self.kind == "code":
            exec codeStr.lstrip() # exec can't handle leading whitespace
            self.code = fun
            self.codeStr = codeStr
        elif self.kind == "data":
            self.code = eval(codeStr.lstrip()) # eval can't handle leading whitespace
        elif self.kind == "primitive":
            self.code = eval(codeStr.lstrip()) # eval can't handle leading whitespace
        return self

    def isEmpty(self):
        return self.code == None
    
    def call(self, container, args):
        if self.kind == "code":
            # set up actual parameters
            for (key,val) in args.items(): # LEGACY: Should use iteritems()
                setterName = "_"+key
                if not container.hasEdge(setterName):
                    raise TubeEvalException, ("unknown parameter",key)
            
                container.dispatchLocal(context=container, edgeName=setterName)(set=val)

            # call
            return self.code(glob=container.environment.glob, me=container)
        elif self.kind == "data":
            return self.code
        elif self.kind == "primitive":
            return self.code(node = container, edge = self.calledEdge, args = args)

    def __repr__(self):
        return repr(self.code)

    def toXML(self):
        if self.isEmpty():
            return ""
        else:
            kind = self.kind
            if self.kind == "code":
                codeStr = self.codeStr
            else:
                codeStr = repr(self.code)
            
            return """
            <content kind="%(kind)s">%(codeStr)s
            </content>""" % vars()


class TubeNode:

    def __init__(self, environment, id=None):
        "Never called externally, only from TubeEnvironment"
        self._edges  = []
        self.environment = environment
        self.content = None
        self.environment.registerNode(self, id)        

    #------------ Three ways of creating a node: by hand, from XML, by copying
    
    def initFromData(self, content):
        self.content = content
        return self

    def initFromXML(self, xmlNode, idDict):
        contentNodes = xmlNode.getElementsByTagName("content")
        if len(contentNodes) > 0 and contentNodes[0].hasChildNodes():
            content = TubeContent().initFromXML(xmlNode=contentNodes[0])
        else:
            content = TubeContent()
        
        self.initFromData(content = content)
        
        # edges
        for edgeNode in xmlNode.getElementsByTagName("edge"):
            edge = TubeEdge().initFromXML(xmlNode=edgeNode, idDict=idDict)
            self.addEdge(edge)
        return self # allows method call chaining

    def genericCopy(self, followEdge, copyDeeply):
        ""
        myCopy = TubeNode(environment = self.environment).initFromData(content = self.content)
        for edge in self._edges:
            if followEdge(edge):
                myCopy.addEdge(edge.genericCopy(followEdge = followEdge,
                                                copyDeeply = copyDeeply))
        return myCopy

    #------------

    def addEdge(self, edge):
        self._edges.append(edge)

    def hasEdge(self, name):
        for edge in self._edges:
            if edge.name == name: return True
        return False
        
    def getEdge(self, name):
        for edge in self._edges:
            if edge.name == name: return edge
        raise IndexError("unknown edge", self._id, name)
        
    def getDelegates(self):
        return [ edge for edge in self._edges if edge.isDelegate() ]
    
    def getParent(self):
        for edge in self._edges:
            if edge.isParent(): return edge
        return None
    
    def hasParent(self):
        return self.getParent() != None
    
    def clone(self):
        "Copy whole subgraph: downward and upward-pointing (unshared) edges"
        return self.genericCopy(followEdge = lambda edge: True,
                                copyDeeply = lambda edge: not edge.isShared())

    def treeCopy(self):
        "Only copy downward-pointing (unshared) edges"
        return self.genericCopy(followEdge = lambda edge: edge.pointsDown(),
                                copyDeeply = lambda edge: not edge.isShared())


    def findMatchAtCurrentLevel(self, edgeName):
        """Searches only this node and its delegates for a match
        (a node that has an edge named 'edgeName')
        
        returns: match or None"""
        if self.hasEdge(edgeName):
            return self
        for delegate in self.getDelegates():
            result = delegate.target.findMatchAtCurrentLevel(edgeName)
            if result != None:
                return result
        return None

    def findMatchAtAnyLevel(self, edgeName):
        """Searches its delegates *and* its ancestors for a match
        (a node that has an edge named 'edgeName')
        
        returns: (context,match) or None"""
        
        match = self.findMatchAtCurrentLevel(edgeName)
        if match != None:
            return (self,match)
        if self.hasParent():
            # We've searched everything on this level, try the next one
            return self.getParent().target.findMatchAtAnyLevel(edgeName)
        else:
            return None

        
    def dispatch(self, edgeName):
        "Returns a callable object"
        result = self.findMatchAtAnyLevel(edgeName)
        if result == None:
            raise AttributeError, ("Could not find match",edgeName)
        (context,match) = result
        return match.dispatchLocal(context=context, edgeName=edgeName)

    def dispatchLocal(self, context, edgeName):
        """Dispatch strictly locally.

        Make a copy of the edge target under 'context' and return it"""
        
        edge = self.getEdge(edgeName)
        targetCopy = edge.target.treeCopy()
        targetCopy.addEdge(TubeEdge().initFromData(target=context, type="parent", shared=True))
        targetCopy.content.calledEdge = edge
        return targetCopy

    def __getattr__(self, name):
        return self.dispatch(name)

## Canonical implementation (template):
##    def __getattr__(self, name):
##        if name == 'x': return 333
##        return object.__getattr__(self, name)

    def __call__(self, **args):
        """context has to be set up before (by copying a node under the context)"""
        return self.content.call(container=self, args=args)

    def toXML(self):
        id = self._id
        contentStr = self.content.toXML()
        edges = ""
        for edge in self._edges:
            edges += edge.toXML()
        
        return """
        <node id="%(id)s">%(contentStr)s
        %(edges)s</node>""" % vars() # a dictionary with all local variables

###################################################################

class PrimSetter:
    # Could have constructor with arguments here!
    
    def __call__(self, node, edge, args):
        newValue = args["set"]
        if not isinstance(newValue, TubeNode):
            newValue = node.environment.makeTubeNode().initFromData(content=TubeContent().initFromData(
                code=newValue, kind="data"))

        changeEdgeName = edge.name[1:] # remove leading underscore

        node.getParent().target.getEdge(changeEdgeName).target = newValue
        return None

    def __repr__(self):
        return "PrimSetter()"

###################################################################

class TubeEnvironment:

    def __init__(self):
        self.tubeNodes = {}
        self.glob = None

    def registerNode(self, node, id = None):
        "Register this node under an ID in tubeNodes"
        
        node._id = id
        if node._id == None:
            idCounter = 1
            while True:
                node._id = "tmpID_" + str(idCounter)
                if not self.tubeNodes.has_key(node._id):
                    # we've found an unused ID
                    break
                idCounter += 1
        self.tubeNodes[node._id] = node

    def makeTubeNode(self, id=None):
        return TubeNode(environment=self, id=id)

    def toXML(self):
        result = '<tubegraph root="%s">' % self.glob._id
        for node in self.tubeNodes.values(): # LEGACY: use itervalues() in newer Python
            result += node.toXML()
        result += '</tubegraph>'
        return result

    def readXML(self, fileName):
        self.tubeNodes = {}
        doc = xml.dom.minidom.parse(fileName)
        # first pass: collect IDs
        for node in doc.documentElement.getElementsByTagName("node"):
            self.makeTubeNode(id=node.getAttribute("id")) # automatically registered in constructor

        # second pass: do the wiring
        for node in doc.documentElement.getElementsByTagName("node"):
            if not self.tubeNodes.has_key(node.getAttribute("id")):
                raise TubeFormatException, ("Unknown ID", node.getAttribute("id"))
                return
            self.tubeNodes[node.getAttribute("id")].initFromXML(
                xmlNode = node, idDict = self.tubeNodes)

        self.glob = self.tubeNodes[doc.documentElement.getAttribute("root")]

    def writeXML(self, fileName):
        file=open(fileName, 'w')
        file.write(toXML())
        file.close()



# from tube import TubeEnvironment; env = TubeEnvironment(); env.readXML("/home/r/rauschma/tube/examples/paper-demo.tube")
# print env.toXML()
# glob.foo(x="haha")
